@@ -80,7 +80,8 @@ char DefaultAltConfFile[] = CONFFILE2;
 char DefaultAltConfDir[] = CONFFILE2 ".d";
 
 enum linetype { Devices, Array, Mailaddr, Mailfrom, Program, CreateDev,
-		Homehost, HomeCluster, AutoMode, Policy, PartPolicy, LTEnd };
+		Homehost, HomeCluster, AutoMode, Policy, PartPolicy, Sysfs,
+		MonitorDelay, LTEnd };
 char *keywords[] = {
 	[Devices]  = "devices",
 	[Array]    = "array",
@@ -93,6 +94,8 @@ char *keywords[] = {
 	[AutoMode] = "auto",
 	[Policy]   = "policy",
 	[PartPolicy]="part-policy",
+	[Sysfs]    = "sysfs",
+	[MonitorDelay] = "monitordelay",
 	[LTEnd]    = NULL
 };
 
@@ -586,6 +589,17 @@ void homeclusterline(char *line)
 	}
 }
 
+static int monitor_delay;
+void monitordelayline(char *line)
+{
+	char *w;
+
+	for (w = dl_next(line); w != line; w = dl_next(w)) {
+		if (monitor_delay == 0)
+			monitor_delay = strtol(w, NULL, 10);
+	}
+}
+
 char auto_yes[] = "yes";
 char auto_no[] = "no";
 char auto_homehost[] = "homehost";
@@ -764,6 +778,12 @@ void conf_file(FILE *f)
 		case PartPolicy:
 			policyline(line, rule_part);
 			break;
+		case Sysfs:
+			sysfsline(line);
+			break;
+		case MonitorDelay:
+			monitordelayline(line);
+			break;
 		default:
 			pr_err("Unknown keyword %s\n", line);
 		}
@@ -920,6 +940,12 @@ char *conf_get_homecluster(void)
 	return home_cluster;
 }
 
+int conf_get_monitor_delay(void)
+{
+	load_conffile();
+	return monitor_delay;
+}
+
 struct createinfo *conf_get_create_info(void)
 {
 	load_conffile();
